#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "lseparable.h"
#include "laminar.h"


static struct lseparable_st *lseparable_inst;
static double ffuncL(int dim, double x[]) { return lseparable(dim, x, lseparable_inst); }
static double ifuncL(int dim, int x[]) { return int_func(dim, ffuncL, x); }


static void read_Lparam(int *init, struct lseparable_st *p) {
    if (!imatrix_scan(p->dim, p->aa)) return;
    if (!imatrix_scan(p->dim, p->bb)) return;
    if (!ivector_scan(p->dim, p->a)) return;
    if (!ivector_scan(p->dim, p->b)) return;
    if (!ivector_scan(p->dim, p->c)) return;
    if (!ivector_scan(p->dim, init)) return;
}

static int v[1000];
static double ffuncM(int dim, double x[]) { return flaminar(dim, x, v); }
static double ifuncM(int dim, int x[]) { return int_func(dim, ffuncM, x); }

static void read_Mparam(const int dim, int *init, int *v, char type) {
    int i, j, k, index, a[100], b[100];

    if (!ivector_scan(dim * 2 - 1, a)) return;
    if (!ivector_scan(dim * 2 - 1, b)) return;
    if (!ivector_scan(dim, init)) return;

    index = 0;
    for (i=1; i/2<dim; i*=2) {
	for (j=0; j+i/2<dim; j+=i) {
            int max = j+i;  if (max > dim) max = dim;
            *v++ = type;
            switch (type) {
            case '2':  /* ax^2 + bx */
                *v++ = a[index];
                *v++ = b[index];
                *v++ = 0; /* c = 0 */
                break;
            case '4':  /* a(x+b)^4 */
                *v++ = a[index] * 1;
                *v++ = a[index] * 4 * b[index];
                *v++ = a[index] * 6 * b[index] * b[index];
                *v++ = a[index] * 4 * b[index] * b[index] * b[index];
                *v++ = a[index] * 1 * b[index] * b[index] * b[index] * b[index];
                break;
            case 'E':  /* a * exp(bx) */
                *v++ = a[index];
                *v++ = b[index];
                break;
            case 'L':  /* -a log( b x ) + x */
                *v++ = a[index];
                *v++ = b[index];
                *v++ = 0;
                break;
            default:
                assert(0 != 0);
                break;
            }
            for (k=j; k<max; k++) *v++ = k;
            *v++ = -1; /* end mark */
            if (type == 'L') {
                *v++ = '2';
                *v++ = 0;
                *v++ = 1;
                *v++ = 0;
                for (k=j; k<max; k++) *v++ = k;
                *v++ = -1; /* end mark */
            }
            index++;
	}
    }
    *v++ = '0'; /* end mark */
}

void demo_Lseparable(const int dim) {
    int init[100];

    lseparable_inst = lseparable_make(dim);
    read_Lparam(init, lseparable_inst);
    /* lseparable_output(lseparable_inst); */

    lgconv_minimize_WOLFE(dim, ifuncL, init);
}

void demo_Mseparable(const int dim, char type) {
    int i;
    int init[100];
    int upper[100];
    int lower[100];

    for (i=0; i<dim; i++) { lower[i] = -100; upper[i] = 100; }
    read_Mparam(dim, init, v, type);
    mgconv_minimize(dim, ifuncM, init, lower, upper);
}
